/**************************************************************************************
Author:       David Szybala
Website:      www.davidszybala.com
Alt Website:  www.twistedgerbil.com
Program:      camera_copy.ms
Date:         11/5/16
Description:  This script copies a scene camera to a free camera with or
              without animation.  Can also mirror scene cameras.
***************************************************************************************/

try (destroyDialog camera_copy_rollout) catch()


struct CameraCopy
(
    scene_cameras = #(),
    display_names = #(),

    /**************************************************************************
    Sorts items by name
    **************************************************************************/
    fn sort_by_name item_one item_two =
    (
        case of
        (
            (item_one.name < item_two.name): -1
            (item_one.name > item_two.name): 1
            default: 0
        ) --end case
    ), --end fn
    
    /**************************************************************************
    Abbreviates long camera names
    **************************************************************************/
    fn collect_display_names =
    (
        for current in this.scene_cameras do
        (
            if current.name.count <= 26 then
                append this.display_names current.name
            else
                append this.display_names ("..." + (substring current.name (current.name.count - 20) current.name.count))
        ) --end for
    ), --end fn

    /**************************************************************************
    collects all scene cameras
    **************************************************************************/
    fn collect_scene_cameras =
    (
        --CRITICAL! Clear previous data
        free this.display_names
        
        this.scene_cameras = for current in cameras where (superClassOf current) == camera collect current

        if this.scene_cameras.count > 0 do
        (
            qsort this.scene_cameras this.sort_by_name
            this.collect_display_names()
        ) --end if
    ), --end fn

    /**************************************************************************
    Clones 'original_camera' and returns the result. Undefined is returned
    if the clone failed
    **************************************************************************/
    fn duplicate_camera original_camera =
    (
        local camera_copy = undefined

        try
        (
            maxOps.CloneNodes original_camera actualNodeList:&original_nodes newNodes:&clone_nodes
            camera_copy = clone_nodes[1]
        ) --end try
        catch()

        camera_copy
    ), --end fn

    /**************************************************************************
    Returns all fov keys of 'current_object'
    **************************************************************************/
    fn get_fov_keys current_object =
    (
        local fov_keys = #()

        if (isProperty current_object "fov") do
        (
            if current_object.fov.controller != undefined do
            (
                for key in current_object.fov.controller.keys do
                    append fov_keys key.time
            ) --end if
        )--end if

        sort fov_keys

        fov_keys --return value
    ), --end fn

    /**************************************************************************
    Copies the transform and fov animation from one camera to another
    **************************************************************************/
    fn track_camera camera_original camera_copy start_frame end_frame =
    (
        local fov_keys = this.get_fov_keys camera_original

        with animate on
        (
            for i = start_frame to end_frame do
                at time i camera_copy.transform = camera_original.transform

            if fov_keys.count > 0 do
            (
                for i = start_frame to end_frame do
                    at time i camera_copy.fov = camera_original.fov
            ) --end if
        ) --end with
    ), --end fn

    /**************************************************************************
    Returns the mirror position of 'current_position' depending on
    the 'mirror_mode'
    **************************************************************************/
    fn get_mirror_position current_position mirror_mode =
    (
        local mirror_position = current_position

        case mirror_mode of
        (
            (1): mirror_position *= [-1, 1, 1]
            (2): mirror_position *= [1, -1, 1]
            (3): mirror_position *= [1, 1, -1]
            (4): mirror_position *= [-1, -1, 1]
            (5): mirror_position *= [1, -1, -1]
            (6): mirror_position *= [-1, 1, -1]
            (7): mirror_position *= [-1, -1, -1]
        )--end case

        mirror_position --return value
    ), --end fn

    /**************************************************************************
    Returns the mirror rotation of 'current_rotation' depending on
    the 'mirror_mode'
    **************************************************************************/
    fn get_mirror_rotation current_rotation mirror_mode =
    (
        local mirror_rotation = current_rotation

        case mirror_mode of
        (
            (1): mirror_rotation = quat mirror_rotation.x -mirror_rotation.y -mirror_rotation.z mirror_rotation.w
            (2): mirror_rotation = quat -mirror_rotation.y -mirror_rotation.x -mirror_rotation.w -mirror_rotation.z
            (3): mirror_rotation = quat mirror_rotation.w mirror_rotation.z mirror_rotation.y mirror_rotation.x
            (4):(
                    mirror_rotation = quat mirror_rotation.x -mirror_rotation.y -mirror_rotation.z mirror_rotation.w
                    mirror_rotation = quat -mirror_rotation.y -mirror_rotation.x -mirror_rotation.w -mirror_rotation.z
                )--end case 4
            (5):(
                    mirror_rotation = quat -mirror_rotation.y -mirror_rotation.x -mirror_rotation.w -mirror_rotation.z
                    mirror_rotation = quat mirror_rotation.w mirror_rotation.z mirror_rotation.y mirror_rotation.x
                )--end case 5
            (6):(
                    mirror_rotation = quat mirror_rotation.w mirror_rotation.z mirror_rotation.y mirror_rotation.x
                    mirror_rotation = quat mirror_rotation.x -mirror_rotation.y -mirror_rotation.z mirror_rotation.w
                )--end case 6
            (7):(
                    mirror_rotation = quat mirror_rotation.x -mirror_rotation.y -mirror_rotation.z mirror_rotation.w
                    mirror_rotation = quat -mirror_rotation.y -mirror_rotation.x -mirror_rotation.w -mirror_rotation.z
                    mirror_rotation = quat mirror_rotation.w mirror_rotation.z mirror_rotation.y mirror_rotation.x
                )--end case 7
        )--end case

        mirror_rotation --return value
    ),--end fn

    /**************************************************************************
    Mirrors the transform of 'camera_object'
    **************************************************************************/
    fn mirror_camera camera_object mirror_mode =
    (
        --CRITICAL! Store position now, rotation will move object
        starting_position = camera_object.position

        if (isProperty camera_object "rotation") do
            camera_object.rotation = this.get_mirror_rotation camera_object.rotation mirror_mode

        camera_object.position =  this.get_mirror_position starting_position mirror_mode

        if (isProperty camera_object.controller "roll_angle") do
            camera_object.controller.roll_angle *= -1
    ), --end fn

    /**************************************************************************
    Copies the transform animation mirrored across an axis
    **************************************************************************/
    fn mirror_camera_animation camera_original camera_mirror start_frame end_frame mirror_mode =
    (
        with animate on
        (
            --CRITICAL! Need a key on all frames before going to function!
            for i = start_frame to end_frame do
                at time i camera_mirror.transform = camera_original.transform

            --Now mirror the camera
            for i = start_frame to end_frame do
                at time i (this.mirror_camera camera_mirror mirror_mode)
        ) --end with
    ) --end fn
) --end struct


(--put script into local scope
    global camera_copy_rollout
    local camera_copy_ini_file = getDir #plugcfg + "\\CameraCopy.ini"
    local camera_copy_rollout_pos

    rollout camera_copy_rollout "Camera Copy"
    (
        local cam_copier = CameraCopy()

        /**********************************************************************
        Enables/disables the copy camera buttons
        **********************************************************************/
        fn camera_dropdown_changed =
        (
            local index = camera_copy_rollout.camera_dropdown.selection

            if cam_copier.scene_cameras.count > 0 do
            (
                if (classOf cam_copier.scene_cameras[index]) == Targetcamera or (classOf cam_copier.scene_cameras[index]) == Physical then
                    camera_copy_rollout.copy_camera_button.enabled = true
                else
                    camera_copy_rollout.copy_camera_button.enabled = false
            ) --end if
        ) --end fn

        /**********************************************************************
        Refreshes the camera dropdown list. Also enables/disables the copy
        and mirror buttons
        **********************************************************************/
        fn refresh_camera_list =
        (
            local button_state = true

            cam_copier.collect_scene_cameras()

            if cam_copier.scene_cameras.count > 0 then
            (
                camera_copy_rollout.camera_dropdown.height = cam_copier.scene_cameras.count * 16
                camera_copy_rollout.camera_dropdown.items = for current in cam_copier.display_names collect current
                camera_copy_rollout.mirror_camera_button.enabled = true
                camera_dropdown_changed()
            )--end if
            else
            (
                camera_copy_rollout.camera_dropdown.height = 16
                camera_copy_rollout.camera_dropdown.items = #("No Cameras Found")
                camera_copy_rollout.copy_camera_button.enabled = camera_copy_rollout.mirror_camera_button.enabled = false
            )--end else
        )--end fn

        /**********************************************************************
        UI creation
        **********************************************************************/
        groupbox camera_group "Camera" pos:[5,5] width:210 height:65
            dropDownList camera_dropdown "" pos:[15,20] width:190 items:#()
            button update_camera_button "Update Camera List" pos:[15,45] width:190 height:20

        groupbox option_group "Options" pos:[5,75] width:210 height:60
            checkbox copy_animation_checkbox "Copy Animation" pos:[15,92] width:150 height:15
            label frame_label "Frame Range:" pos:[15,115] width:75 height:15 enabled:false
            spinner frame_start_spinner "" pos:[90,115] width:50 height:15 range:[-99999,99999,0] type:#integer enabled:false
            label dash_label "-" pos:[145,115] width:5 height:15 enabled:false
            spinner frame_end_spinner "" pos:[155,115] width:50 height:15 range:[-99999,99999,100] type:#integer enabled:false

        groupbox copy_group "Copy" pos:[5,140] width:210 height:45
            button copy_camera_button "Copy to Free Camera" pos:[15,157] width:190 height:20 enabled:false

        groupbox mirror_group "Mirror" pos:[5,190] width:210 height:95
            label mirror_label "Axis:" pos:[15,210] width:35 height:15
            radioButtons mirror_axis_radio "" pos:[55,210] width:80 height:35 labels:#("X", "Y", "Z","XY","YZ","ZX","XYZ") default:2 columns:3
            button mirror_camera_button "Mirror Camera" pos:[15,260] width:190 height:20 enabled:false

        label status_label "INFO: Standing by..." pos:[5,295] width:210 height:15

        /**********************************************************************
        UI open event handler. Prepares the UI for first time open
        **********************************************************************/
        on camera_copy_rollout open do
            refresh_camera_list()

        /**********************************************************************
        UI move event handler. Stores the curent position of the UI
        **********************************************************************/
        on camera_copy_rollout moved pos do
            setIniSetting camera_copy_ini_file "Dialog" "Position" (pos as string)

        /**********************************************************************
        Camera dropdown changed event handler. Enables/disables the copy button
        depending on the class of the selected camera
        **********************************************************************/
        on camera_dropdown selected index do
            camera_dropdown_changed()

        /**********************************************************************
        Update camera event handler. Updates the camera dropdown list
        **********************************************************************/
        on update_camera_button pressed do
        (
            refresh_camera_list()
            status_label.text = "INFO: Camera list updated"
        ) --end on

        /**********************************************************************
        Copy animation checkbox event handler. Enables/disables the key all
        frames checkbox depending on the current state
        **********************************************************************/
        on copy_animation_checkbox changed state do
            frame_label.enabled = frame_start_spinner.enabled = dash_label.enabled = frame_end_spinner.enabled = state

        /**********************************************************************
        Start frame changed event handler. Ensures the start frame is not greater than the end frame
        **********************************************************************/
        on frame_start_spinner changed the_value do
        (
            if the_value > frame_end_spinner.value do
                frame_end_spinner.value = the_value
        )--end on

        /**********************************************************************
        End frame changed event handler. Ensures the end frame is not less than the start frame
        **********************************************************************/
        on frame_end_spinner changed the_value do
        (
            if the_value < frame_start_spinner.value do
                frame_start_spinner.value = the_value
        )--end on

        /**********************************************************************
        Copy camera event handler. Copies the selected camera to a free camera.
        NOTE: Only targeted and physical cameras can be copied at this time
        **********************************************************************/
        on copy_camera_button pressed do
        (
            local camera_original = cam_copier.scene_cameras[camera_dropdown.selection]

            if isDeleted camera_original do
            (
                status_label.text = "ERROR: Camera does not exist. Update list"
                return false
            ) --end if

            local camera_copy = Freecamera name:(uniqueName (camera_original.name + "_Copy_"))
            camera_copy.transform = camera_original.transform
            camera_copy.fov = camera_original.fov

            if copy_animation_checkbox.checked do
                cam_copier.track_camera camera_original camera_copy frame_start_spinner.value frame_end_spinner.value

            refresh_camera_list()
            status_label.text = "INFO: Free camera created"
        )--end on

        /**********************************************************************
        Mirror camera event handler. Mirrors the selected camera
        **********************************************************************/
        on mirror_camera_button pressed do
        (
            local camera_original = cam_copier.scene_cameras[camera_dropdown.selection]

            if (isDeleted camera_original) do
            (
                status_label.text = "ERROR: Camera does not exist. Update list"
                return false
            ) --end if

            local camera_mirror = cam_copier.duplicate_camera camera_original

            if camera_mirror == undefined do
            (
                status_label.text = "ERROR: Could not duplicate camera"
                return false
            ) --end if

            case mirror_axis_radio.state of
            (
                (1): camera_mirror.name = uniqueName (camera_original.name + "_MirrorX_")
                (2): camera_mirror.name = uniqueName (camera_original.name + "_MirrorY_")
                (3): camera_mirror.name = uniqueName (camera_original.name + "_MirrorZ_")
                (4): camera_mirror.name = uniqueName (camera_original.name + "_MirrorXY_")
                (5): camera_mirror.name = uniqueName (camera_original.name + "_MirrorYZ_")
                (6): camera_mirror.name = uniqueName (camera_original.name + "_MirrorZX_")
                (7): camera_mirror.name = uniqueName (camera_original.name + "_MirrorXYZ_")
            )--end case

            if copy_animation_checkbox.checked then
            (
                cam_copier.mirror_camera_animation camera_original camera_mirror frame_start_spinner.value frame_end_spinner.value mirror_axis_radio.state

                if camera_mirror.target != undefined do
                    cam_copier.mirror_camera_animation camera_original.target camera_mirror.target frame_start_spinner.value frame_end_spinner.value mirror_axis_radio.state
            ) --end if
            else
            (
                if camera_mirror.target != undefined do
                (
                    deleteKeys camera_mirror.target #allKeys
                    cam_copier.mirror_camera camera_mirror.target mirror_axis_radio.state
                ) --end if

                deleteKeys camera_mirror #allKeys
                cam_copier.mirror_camera camera_mirror mirror_axis_radio.state
            ) --end else

            refresh_camera_list()
            status_label.text = "INFO: Mirror camera created"
        )--end on
    )--end rollout

    camera_copy_rolloutPos = execute (getIniSetting camera_copy_ini_file "Dialog" "Position")
    if camera_copy_rolloutPos == OK do camera_copy_rolloutPos = [100,100]
    if camera_copy_rolloutPos.x > sysinfo.desktopsize.x - 220 do
        camera_copy_rolloutPos.x = sysinfo.desktopsize.x - 225
    if camera_copy_rolloutPos.y > sysinfo.desktopsize.y - 370 do
        camera_copy_rolloutPos.y = sysinfo.desktopsize.y - 370
    if camera_copy_rolloutPos.x < 0 do camera_copy_rolloutPos.x = 0
    if camera_copy_rolloutPos.y < 0 do camera_copy_rolloutPos.y = 0
    createDialog camera_copy_rollout 220 310 camera_copy_rolloutPos.x camera_copy_rolloutPos.y
)--end local scope
